# Copyright 2023 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Specifies Task to store excclusion specific information."""

from __future__ import annotations

import dataclasses
import datetime
import enum
import math
import uuid
from typing import Final

from croniter import croniter

from googleads_housekeeper.services import enums

_HOURS_IN_DAY: Final[int] = 24


class TaskStatus(enum.Enum):
  """An enumeration representing the status of a task.

  Attributes:
    ACTIVE (int): Represents an active task.
    INACTIVE (int): Represents an inactive task.
  """

  ACTIVE = 0
  INACTIVE = 1


class TaskOutput(enum.Enum):
  """An enumeration representing the artifact/output options for a task.

  Attributes:
    NOTIFY: Output option to notify about the task.
    EXCLUDE: Output option to exclude the task.
    EXCLUDE_AND_NOTIFY: Output option to exclude the task and notify about it.
  """

  NOTIFY = 1
  EXCLUDE = 2
  EXCLUDE_AND_NOTIFY = 3


@dataclasses.dataclass
class OldTask:
  """Defines old_task that needs to be run by an application."""

  _id: int
  customer_id: str
  date_created: str
  email_alerts: bool
  from_days_ago: str
  gads_data_display: bool
  gads_data_youtube: bool
  gads_filter: str
  include_youtube: bool
  lookback_days: str
  schedule: str
  task_id: str
  task_name: str
  yt_country_operator: str
  yt_country_value: str
  yt_language_operator: str
  yt_language_value: str
  yt_std_character: str
  yt_subscriber_operator: str
  yt_subscriber_value: str
  yt_video_operator: str
  yt_video_value: str
  yt_view_operator: str
  yt_view_value: str

  def __post_init__(self) -> None:
    """Ensures safe casting to proper enums."""
    self.__name__ = 'tasks'

  def to_dict(self) -> dict[str, int | str | bool]:
    """Converts class to dictionary.

    Enum attributes of the class are represented as names.
    """
    task_dict = {}
    for key, value in dataclasses.asdict(self).items():
      if hasattr(value, 'value'):
        task_dict[key] = value.name
      else:
        task_dict[key] = value
    return task_dict


@dataclasses.dataclass
class Task:
  """Defines task that needs to be run by an application.

  Task contains all necessary information to find the appropriate placements
  that should be excluded and perform the exclusion.

  Attributes:
    name: Task name.
    exclusion_rule: String version of exclusion rule.
    date_range: Lookback days for data fetching from Google Ads API.
    from_days_ago: start_date of data fetching from Google Ads API.
    exclusion_level: Level of exclusion (AD_GROUP, CAMPAIGN, etc.).
    output: Desired action (EXCLUDE, NOTIFY, etc.)
    status: Task status.
    schedule: Optional task schedule.
    creation_date: Time when task was created.
    id: Unique task identifier.
  """

  name: str
  exclusion_rule: str
  customer_ids: str
  date_range: int = 7
  from_days_ago: int = 0
  exclusion_level: enums.ExclusionLevelEnum = enums.ExclusionLevelEnum.AD_GROUP
  placement_types: str | None = None
  output: TaskOutput = TaskOutput.EXCLUDE_AND_NOTIFY
  status: TaskStatus = TaskStatus.ACTIVE
  schedule: str = '0'
  creation_date: datetime.datetime = datetime.datetime.now()
  id: str = dataclasses.field(default_factory=lambda: str(uuid.uuid4()))

  def __post_init__(self) -> None:
    """Ensures safe casting to proper enums."""
    self.output = self._cast_to_enum(TaskOutput, self.output)
    self.status = self._cast_to_enum(TaskStatus, self.status)
    self.exclusion_level = self._cast_to_enum(
      enums.ExclusionLevelEnum, self.exclusion_level
    )

  def to_dict(self) -> dict[str, int | str]:
    """Converts class to dictionary.

    Enum attributes of the class are represented as names.
    """
    task_dict = {}
    for key, value in dataclasses.asdict(self).items():
      if hasattr(value, 'value'):
        task_dict[key] = value.name
      else:
        task_dict[key] = value
    return task_dict

  @property
  def cron_schedule(self) -> str | None:
    """Builds cron schedule based on creation time and schedule."""
    minute = self.creation_date.strftime('%M')
    hour = self.creation_date.strftime('%H')

    if not (schedule := self.schedule) or self.schedule == '0':
      return None
    if 0 < int(schedule) < _HOURS_IN_DAY:
      schedule = f'{minute} */{schedule} * * *'
    elif int(schedule) >= _HOURS_IN_DAY:
      days = math.floor(int(self.schedule) / _HOURS_IN_DAY)
      schedule = f'{minute} {hour} */{days} * *'
    return schedule

  @property
  def start_date(self) -> str:
    """Formatted start_date."""
    return (
      (
        datetime.datetime.now()
        - datetime.timedelta(days=int(self.from_days_ago + self.date_range))
      )
      .date()
      .strftime('%Y-%m-%d')
    )

  @property
  def end_date(self) -> str:
    """Formatted end_date."""
    return (
      (
        datetime.datetime.now()
        - datetime.timedelta(days=int(self.from_days_ago))
      )
      .date()
      .strftime('%Y-%m-%d')
    )

  @property
  def accounts(self) -> list[str]:
    """Formatted account list."""
    return (
      self.customer_ids.split(',')
      if isinstance(self.customer_ids, str)
      else self.customer_ids
    )

  @property
  def next_run(self) -> str:
    """Next task run as a cron expression."""
    if not (schedule := self.cron_schedule):
      return 'Not scheduled'
    return (
      croniter(schedule, datetime.datetime.now())
      .get_next(datetime.datetime)
      .strftime('%Y-%m-%d %H:%M')
    )

  def _cast_to_enum(
    self, enum_: type[enum.Enum], value: str | enum.Enum
  ) -> enum.Enum:
    """Helper method for converted in string to desired enum.

    Args:
      enum_: Enum class to server as conversion base.
      value: Value to perform the conversion.

    Returns:
      Correct enum based on provided value.
    """
    return enum_[value] if isinstance(value, str) else value



# Copyright 2023 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Injects dependencies to run the application."""

# pylint: disable=C0330, g-bad-import-order g-multiple-import, g-import-not-at-top

from __future__ import annotations

import importlib
import inspect
import os

from gaarf.api_clients import GoogleAdsApiClient

if os.environ.get(
  'ADS_HOUSEKEEPER_DEPLOYMENT_TYPE'
) == 'Google Cloud' and not importlib.util.find_spec('google.cloud.firestore'):
  raise ImportError(
    'Please install googleads-housekeeper with Google Cloud support - '
    '`pip install googleads-housekeeper[gcp]`'
  )
from google.cloud import firestore, pubsub_v1

from googleads_housekeeper.adapters import notifications, orm, publisher
from googleads_housekeeper.services import handlers, messagebus, unit_of_work


class GoogleCloudBootstrapper:
  """Prepares dependencies to run application on Google Cloud."""

  def __init__(
    self,
    topic_prefix: str | None = None,
    database_uri: str = os.environ.get('FIRESTORE_DB', '(default)'),
    project_id: str = os.environ.get('GOOGLE_CLOUD_PROJECT'),
    appengine_region: str = os.environ.get('APPENGINE_REGION', 'europe-west1'),
  ) -> None:
    self.start_orm = False
    self.uow = unit_of_work.FirestoreUnitOfWork(
      firestore.Client(project=project_id, database=database_uri)
    )
    self.notification_service = (
      notifications.GoogleCloudAppEngineEmailNotifications(
        project_id=project_id, appengine_region=appengine_region
      )
    )
    self.publish_service = publisher.GoogleCloudPubSubPublisher(
      client=pubsub_v1.PublisherClient(),
      project_id=project_id,
      topic_prefix=topic_prefix,
    )


class LocalBootstrapper:
  """Prepares dependencies to run application locally."""

  def __init__(
    self, topic_prefix: str | None = None, database_uri: str | None = None
  ) -> None:
    self.start_orm = True
    self.uow = unit_of_work.SqlAlchemyUnitOfWork(database_uri or _get_db_uri())
    self.publish_service = publisher.LogPublisher(topic_prefix)
    self.notification_service = notifications.ConsoleNotifications()


class DevBootstrapper:
  """Prepares dependencies to run application in development environment."""

  def __init__(
    self, topic_prefix: str | None = None, database_uri: str | None = None
  ) -> None:
    self.start_orm = True
    self.uow = unit_of_work.SqlAlchemyUnitOfWork(
      database_uri or f'sqlite:///{os.getcwd()}/ads_housekeeper_tasks.db'
    )
    self.publish_service = publisher.LogPublisher(topic_prefix)
    self.notification_service = notifications.ConsoleNotifications()


DEPLOYMENT_TYPES = {
  'Local': LocalBootstrapper,
  'Google Cloud': GoogleCloudBootstrapper,
  'Dev': DevBootstrapper,
}


class Bootstrapper:
  """Prepares application dependencies based on a provided deployment type."""

  def __init__(
    self,
    deployment_type: str = 'Dev',
    topic_prefix: str | None = None,
    database_uri: str | None = None,
    **kwargs: str,
  ) -> None:
    self.deployment_type = deployment_type
    self.topic_prefix = topic_prefix
    self.database_uri = database_uri
    self.bootstrapper_parameters = kwargs or {}

  def bootstrap_app(
    self,
  ) -> messagebus.MessageBus:
    """Builds MessageBus based on specified deployment type."""
    bootstrapper = DEPLOYMENT_TYPES[self.deployment_type](
      topic_prefix=self.topic_prefix,
      database_uri=self.database_uri,
    )
    return bootstrap(
      start_orm=bootstrapper.start_orm,
      uow=bootstrapper.uow,
      notification_service=bootstrapper.notification_service,
      publish_service=bootstrapper.publish_service,
    )


def _get_db_uri(
  host: str = os.environ.get('DB_HOST', 'localhost'),
  port: int = int(os.environ.get('DB_PORT', 54321)),
  user: str = os.environ.get('DB_USER', 'ads_housekeeper'),
  password: str = os.environ.get('DB_PASSWORD', 'ads_housekeeper'),
  db_name: str = os.environ.get('DB_NAME', 'ads_housekeeper'),
  db_engine: str = os.environ.get('DB_ENGINE', 'postgresql'),
) -> str:
  if dev_db := os.environ.get('DEV_DB_HOST'):
    return f'sqlite:///{dev_db}'
  return f'{db_engine}://{user}:{password}@{host}:{port}/{db_name}'


def _get_default_google_ads_api_client():
  return GoogleAdsApiClient(
    path_to_config=os.environ.get(
      'GOOGLE_ADS_PATH_TO_CONFIG', '~/google-ads.yaml'
    ),
    version=os.environ.get('GOOGLE_ADS_API_VERSION', 'v16'),
  )


def _get_notification_service() -> notifications.BaseNotifications:
  notification_service_factory = notifications.NotificationFactory()
  notification_type = os.environ.get(
    'ADS_HOUSEKEEPER_NOTIFICATION_TYPE', 'console'
  )
  return notification_service_factory.create_notification_service(
    notification_type=notification_type
  )


def bootstrap(
  ads_api_client: GoogleAdsApiClient | None = None,
  start_orm: bool = True,
  uow: unit_of_work.AbstractUnitOfWork = unit_of_work.SqlAlchemyUnitOfWork(
    _get_db_uri()
  ),
  notification_service: notifications.BaseNotifications = (
    _get_notification_service()
  ),
  publish_service: publisher.BasePublisher = publisher.LogPublisher(),
) -> messagebus.MessageBus:
  """Injects dependencies to run the application.

  Args:
    ads_api_client: Google Ads API client to fetch data and perform mutates.
    start_orm: Whether to initialize ORM object in DB.
    uow: Unit of Work to handle transactions.
    notification_service: Service to send notifications on operations retults.
    publish_service: Service to send events.

  Returns:
    MessageBus with injected dependencies.
  """
  if start_orm:
    orm.start_mappers(engine=uow.engine)

  dependencies = {
    'uow': uow,
    'ads_api_client': ads_api_client or _get_default_google_ads_api_client(),
    'notification_service': notification_service,
    'publisher': publish_service,
  }
  injected_event_handlers = {
    event_type: [
      _inject_dependencies(handler, dependencies) for handler in handlers
    ]
    for event_type, handlers in handlers.EVENT_HANDLERS.items()
  }
  injected_command_handlers = {
    command_type: _inject_dependencies(handler, dependencies)
    for command_type, handler in handlers.COMMAND_HANDLERS.items()
  }
  return messagebus.MessageBus(
    uow=uow,
    event_handlers=injected_event_handlers,
    command_handlers=injected_command_handlers,
    dependencies=dependencies,
    is_observe_mode=os.getenv('ADS_HOUSEKEEPER_MODE', 'Observe').lower()
    == 'observe',
  )


def _inject_dependencies(handler, dependencies):
  params = inspect.signature(handler).parameters
  deps = {
    name: dependency
    for name, dependency in dependencies.items()
    if name in params
  }
  return lambda message: handler(message, **deps)



# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Specifies various parsers to get data from non Google Ads API."""

# pylint: disable=C0330, g-bad-import-order, g-multiple-import

from collections.abc import Sequence
from typing import Callable

import garf_core
import garf_youtube_data_api

from googleads_housekeeper.domain.external_parsers import website_parser


def get_video_info(
  placements: Sequence[str], **kwargs: str
) -> dict[str, garf_core.report.GarfRow]:
  """Calls YouTube Data API to get information on provided YouTube videos."""
  videos_query = """
  SELECT
    id,
    snippet.title AS title,
    snippet.description AS description,
    snippet.defaultLanguage AS defaultLanguage,
    snippet.defaultAudioLanguage AS defaultAudioLanguage,
    statistics.commentCount AS commentCount,
    statistics.favouriteCount AS favouriteCount,
    statistics.likeCount AS likeCount,
    statistics.viewCount AS viewCount,
    status.madeForKids AS madeForKids,
    snippet.tags AS tags,
    topicDetails.topicCategories AS topicCategories
  FROM channels
  """

  entities = []
  for row in placements:
    for key, value in kwargs.items():
      if row[key] == value:
        entities.append(row.placement)
  if not entities:
    return {}
  results = garf_youtube_data_api.YouTubeDataApiReportFetcher(
    garf_youtube_data_api.YouTubeDataApiClient()
  ).fetch(videos_query, id=placements)
  output = {}
  for row in results:
    output[row.id] = row
  return output


def get_channel_info(
  placements: Sequence[str], **kwargs: str
) -> dict[str, garf_core.report.GarfRow]:
  """Calls YouTube Data API to get information on provided YouTube videos."""
  channels_query = """
  SELECT
    id,
    snippet.title AS title,
    snippet.description AS description,
    statistics.subscriberCount AS subscriberCount,
    statistics.videoCount AS videoCount,
    statistics.viewCount AS viewCount,
    status.madeForKids AS madeForKids,
    topicDetails.topicCategories AS topicCategories
  FROM channels
  """
  entities = []
  for row in placements:
    for key, value in kwargs.items():
      if row[key] == value:
        entities.append(row.placement)
  if not entities:
    return {}
  results = garf_youtube_data_api.YouTubeDataApiReportFetcher(
    garf_youtube_data_api.YouTubeDataApiClient()
  ).fetch(channels_query, id=entities)
  output = {}
  for row in results:
    output[row.id] = row
  return output


def get_website_info(
  placements: Sequence[str], **kwargs: str
) -> dict[str, garf_core.report.GarfRow]:
  """Performs web scraping of provided website urls."""
  entities = []
  for row in placements:
    for key, value in kwargs.items():
      if row[key] == value:
        entities.append(row.placement)
  if not entities:
    return {}
  results = website_parser.WebSiteParser().parse(entities)
  output = {}
  for row in results:
    output[row.placement] = garf_core.report.GarfRow(
      data=[row.title, row.description, row.keywords],
      column_names=['title', 'description', 'keywords'],
    )
  return output


PARSER_MAPPING: dict[
  str, Callable[[Sequence[str]], dict[str, garf_core.report.GarfRow]]
] = {
  'YOUTUBE_CHANNEL_INFO': get_channel_info,
  'YOUTUBE_VIDEO_INFO': get_video_info,
  'WEBSITE_INFO': get_website_info,
}



# Copyright 2023 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Module for defining various notification channels."""

from __future__ import annotations

import abc
import dataclasses
import os

import gaarf
import slack_sdk
from google.appengine.api import mail

from googleads_housekeeper.domain.core.task import TaskOutput


@dataclasses.dataclass
class MessagePayload:
  """Contains information to be sent via a given notification channel.

  Attributes:
      task_id:
          task identifier.
      task_name:
          task name.
      task_formula:
          task full exclusion_rule.
      task_output:
          task action verb.
      placements_excluded_sample:
          optional report with top N placements that were excluded.
      total_placement_excluded:
          number of all excluded placements by a given task.
      recipient:
          who should receive the notification (could be email, slack channel).
  """

  task_id: str
  task_name: str
  task_formula: str
  task_output: TaskOutput
  placements_excluded_sample: gaarf.report.GaarfReport | None
  total_placement_excluded: int
  recipient: str


class BaseNotifications(abc.ABC):
  """Base class for sending notifications."""

  @abc.abstractmethod
  def send(self, payload: MessagePayload) -> None:
    """Transforms payload and sends it via a given notification channel."""
    raise NotImplementedError


def get_region_initials(region: str):
  """Reads the region value and returns its initials.

  Args:
      region (str): The region's name.

  Returns:
      str: The initials of the region value.
  """
  return ''.join(word[0] for word in region.lower().split('-'))


class GoogleCloudAppEngineEmailNotifications(BaseNotifications):
  """Implements sending emails with App Engine.

  Attributes:
      project_id: Google Cloud project id.
  """

  def __init__(self, project_id: str, appengine_region: str) -> None:
    """Initializes GoogleCloudAppEngineEmailNotifications class.

    Args:
        project_id: Id of Google Cloud Project.
        appengine_region: Appengine's Region.
    """
    self.project_id = project_id
    self.appengine_region_initials = get_region_initials(appengine_region)

  def send(self, payload: MessagePayload, **kwargs: str) -> None:
    """Sends payload to via appengine mailer."""
    message = self._prepare_message(payload)
    message.send()

  def _prepare_message(self, payload: MessagePayload) -> mail.Message:
    """Generates message from a payload."""
    sender = f'exclusions@{self.project_id}.appspotmail.com'
    if placements_excluded := payload.placements_excluded_sample:
      output = payload.task_output
      subject = (
        f"{'Detected' if output == TaskOutput.NOTIFY else 'Excluded'} "
        f'{payload.total_placement_excluded} placements'
      )
      body = self._generate_notification_html(
        payload.task_id,
        payload.task_name,
        payload.task_formula,
        payload.task_output,
        placements_excluded,
      )
    else:
      subject = 'No placements were detected'
      body = ''
    recipient = payload.recipient.split(',')
    if body:
      return mail.EmailMessage(
        sender=sender, to=recipient, subject=subject, html=body
      )
    return mail.EmailMessage(
      sender=sender, to=recipient, subject=subject, body=body
    )

  def _generate_notification_html(
    self,
    task_id: str,
    task_name: str,
    task_formula: str,
    task_action_verb: TaskOutput,
    placements: gaarf.report.GaarfReport,
  ) -> str:
    html_body = """
        <!DOCTYPE html>
        <html lang="en">
        <head>
            <meta charset="UTF-8">
            <meta name="viewport" content="width=device-width,
            initial-scale=1.0">
            <title>CPR Placement Exclusion Alert</title>
            <style type="text/css">
        body {
            font-family: Arial, sans-serif;
            margin: 0;
            padding: 20px;
            background-color: #f8f8f8; /* Light background for visual
            separation */
        }

        .container {
            background-color: #fff; /* White background for content */
            border: 1px solid #ddd;
            padding: 20px;
            border-radius: 8px;
            box-shadow: 0 2px 4px rgba(0,0,0,0.1); /* Subtle shadow for depth */
        }

        h2 {
            color: #2c3e50; /* A darker heading color */
            margin-bottom: 15px;
            font-weight: bold;
        }

        .task-link {
            color: #007bff;
            text-decoration: none;
            font-weight: 600; /* Slightly bolder task link */
        }

        .placement-list {
            list-style: disc;
            margin-left: 20px;
        }

        .logo-placeholder {
            text-align: center;
            margin-top: 20px;
        }

        .logo-placeholder img {
        max-width: 150px; /* Limit logo size */
        }

            </style>
        </head>
        <body>
            <p>Please do not reply to this email.</p>

            <div class="container">
                <h2>CPR Placement Exclusion Alert</h2>"""

    html_body += f"""
                <p>The CPR tool has {("detected" if task_action_verb == TaskOutput.NOTIFY
                                                               else "excluded")} the following placements
                based on your defined thresholds:</p>
                <p><strong>Task Name: </strong>
                <a href="https://{self.project_id}.
                {self.appengine_region_initials}.r.appspot.com/newtask?task={task_id}"
                class="task-link" target="_blank" title="View Task Details">
                {task_name}</a></p>
                <p><strong>Task Logic: </strong>{_to_human_readable(task_formula)}
                </p>"""

    for placement in placements:
      placement_name = placement.get('name')
      placement_url = placement.get('placement')
      placement_url_html = (
        '' if placement_url == placement_name else f'<br/>{placement_url}'
      )
      html_body += f"""<ul class="placement-list">
            <strong>Placement Name:</strong> {placement_name}
            {placement_url_html}
            <br/>
            Reason: {_to_human_readable(getattr(placement, 'reason',
                                                'Task without filters'))}</li>

        </ul>
                  """
    html_body += """
            </div>

            <div class="logo-placeholder">
                <img src="https://services.google.com/fh/files/emails/gtech_logo_transparent.png"
                alt="gPS Ads Logo" width="32" height="32">
            </div>

            <p>Please do not reply to this email.</p>
        </body>
        </html>
        """
    return html_body


def _to_human_readable(text: str) -> str:
  return text.lower().replace('_', ' ')


class SlackNotifications(BaseNotifications):
  """Implements sending emails with Slack.

  Attributes:
      bot_token: Slack API bot token.
      client: initialized Slack WebClient.
  """

  def __init__(self, bot_token: str) -> None:
    """Initializes SlackNotifications class.

    Args:
        bot_token: Slack API Token.
    """
    self.client = slack_sdk.WebClient(token=bot_token)

  def send(self, payload: MessagePayload) -> None:
    """Sends payload to Slack channel."""
    task_name = payload.task_name
    channel = payload.recipient
    if not (placements_excluded_sample := payload.placements_excluded_sample):
      self.client.chat_postMessage(
        channel=channel, text=f'{task_name}: No placements were excluded'
      )
    else:
      file = (
        placements_excluded_sample['name']
        .to_pandas()
        .to_csv(index=False, sep='\t')
      )
      self.client.files_upload(
        channels=channel,
        initial_comment=(
          f'{task_name}: '
          f'{payload.total_placement_excluded} placements excluded'
        ),
        filename=f'{task_name}.tsv' if task_name else 'cpr.tsv',
        content=file,
      )


class ConsoleNotifications(BaseNotifications):
  """Implements sending notification to standard output."""

  def send(self, payload: MessagePayload) -> None:
    """Prints payload to standard output."""
    print(payload)


class NullNotifications(BaseNotifications):
  """Implements special case for incorrect notification channel."""

  def __init__(self, notification_type: str) -> None:
    self.notification_type = notification_type

  def send(self, payload: MessagePayload) -> None:
    raise ValueError(
      f'Cannot send notification via '
      f'"{self.notification_type}" notification channel'
    )


class NotificationFactory:
  """Builds notification service for a given notification type.

  NotificationFactory contains all possible types of notification channels
  defined in the module.

  Attributes:
      types: mapping between name of services and its class.

  """

  types: dict[str, dict[str, BaseNotifications | dict]] = {}

  def __init__(self) -> None:
    self.load_types()

  def load_types(self) -> None:
    self.types['email'] = {
      'type': GoogleCloudAppEngineEmailNotifications,
      'args': {
        'project_id': os.environ.get('GOOGLE_CLOUD_PROJECT'),
        'appengine_region': os.environ.get('APPENGINE_REGION', 'europe-west1'),
      },
    }
    self.types['slack'] = {
      'type': SlackNotifications,
      'args': {'bot_token': os.environ.get('CPR_SLACK_BOT_TOKEN')},
    }
    self.types['console'] = {'type': ConsoleNotifications, 'args': {}}

  def create_notification_service(
    self, notification_type: str
  ) -> type[BaseNotifications]:
    if notification_type in self.types:
      if args := self.types[notification_type].get('args'):
        return self.types[notification_type].get('type')(**args)
      return self.types[notification_type].get('type')()
    return NullNotifications(notification_type)



from __future__ import annotations

from fastapi import APIRouter, Body

from googleads_housekeeper import views
from googleads_housekeeper.domain import commands
from googleads_housekeeper.entrypoints.bus import bus

router = APIRouter(prefix='/settings')


@router.get('/')
def list_config() -> list[dict]:
  return views.config(bus.uow)


@router.post('/')
def create_config(data: dict = Body(embed=True)):
  return bus.handle(commands.SaveConfig(**data))



# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Defines ORM mapping between entities and DB tables."""

# pylint: disable=C0330

from __future__ import annotations

from sqlalchemy import (
  Boolean,
  Column,
  DateTime,
  Enum,
  ForeignKey,
  Integer,
  MetaData,
  String,
  Table,
)
from sqlalchemy.orm import mapper, relationship

from googleads_housekeeper.domain.core import (
  execution,
  settings,
  task,
)
from googleads_housekeeper.domain.placement_handler import (
  entities as placement_entities,
)
from googleads_housekeeper.services import enums

metadata = MetaData()

tasks = Table(
  'tasks',
  metadata,
  Column('_id', Integer, autoincrement=True),
  Column('id', String(36), primary_key=True),
  Column('name', String(255)),
  Column('exclusion_rule', String(255)),
  Column('customer_ids', String(255)),
  Column('date_range', Integer),
  Column('exclusion_level', Enum(enums.ExclusionLevelEnum)),
  Column('output', Enum(task.TaskOutput)),
  Column('from_days_ago', Integer),
  Column('placement_types', String(255)),
  Column('creation_date', DateTime),
  Column('status', Enum(task.TaskStatus)),
  Column('schedule', String(255)),
)

old_tasks = Table(
  'old_tasks',
  metadata,
  Column('_id', Integer, autoincrement=True),
  Column('customer_id', String(36)),
  Column('date_created', String(255)),
  Column('email_alerts', Boolean),
  Column('from_days_ago', String(36)),
  Column('gads_data_display', Boolean),
  Column('gads_data_youtube', Boolean),
  Column('gads_filter', String(255)),
  Column('include_youtube', Boolean),
  Column('lookback_days', String(36)),
  Column('schedule', String(36)),
  Column('task_id', String(36), primary_key=True),
  Column('task_name', String(255)),
  Column('yt_country_operator', String(255)),
  Column('yt_country_value', String(255)),
  Column('yt_language_operator', String(255)),
  Column('yt_language_value', String(255)),
  Column('yt_std_character', String(255)),
  Column('yt_subscriber_operator', String(255)),
  Column('yt_subscriber_value', String(255)),
  Column('yt_video_operator', String(255)),
  Column('yt_video_value', String(255)),
  Column('yt_view_operator', String(255)),
  Column('yt_view_value', String(255)),
)

executions = Table(
  'executions',
  metadata,
  Column('_id', Integer, autoincrement=True),
  Column('id', String(36), primary_key=True, nullable=False),
  Column('task', ForeignKey('tasks.id'), nullable=False),
  Column('start_time', DateTime, nullable=False),
  Column('end_time', DateTime, nullable=False),
  Column('type', Enum(execution.ExecutionTypeEnum)),
  Column('placements_excluded', Integer),
)

execution_details = Table(
  'execution_details',
  metadata,
  Column('_id', Integer, autoincrement=True),
  Column('id', String(36), primary_key=True, nullable=False),
  Column('execution_id', ForeignKey('executions.id'), nullable=False),
  Column('placement', String(255), nullable=False),
  Column('placement_type', String(255), nullable=False),
  Column('reason', String(255), nullable=False),
)


allowlisted_placements = Table(
  'allowlisted_placements',
  metadata,
  Column('_id', Integer, autoincrement=True),
  Column('id', String(36), primary_key=True, nullable=False),
  Column('type', Enum(enums.PlacementTypeEnum)),
  Column('name', String(255), unique=True),
  Column('account_id', String(255)),
)

config = Table(
  'config',
  metadata,
  Column('_id', Integer, autoincrement=True),
  Column('id', String(36), primary_key=True, nullable=False),
  Column('mcc_id', String(10)),
  Column('always_fetch_youtube_preview_mode', Boolean),
  Column('save_to_db', Boolean),
  Column('email_address', String(50)),
)

customer_ids = Table(
  'customer_ids',
  metadata,
  Column('mcc_id', String(10), primary_key=True),
  Column('account_name', String(255)),
  Column('id', String(10), primary_key=True),
)

mcc_ids = Table(
  'mcc_ids',
  metadata,
  Column('id', String(10), primary_key=True),
  Column('account_name', String(255)),
)


def start_mappers(engine):
  execution_details_mapper = mapper(
    execution.ExecutionDetails, execution_details
  )
  execution_mapper = mapper(
    execution.Execution,
    executions,
    properties={
      'details': relationship(execution_details_mapper, collection_class=list)
    },
  )
  mapper(
    task.Task,
    tasks,
    properties={
      'executions': relationship(execution_mapper, collection_class=list)
    },
  )
  mapper(task.OldTask, old_tasks)
  mapper(settings.Config, config)
  mapper(placement_entities.AllowlistedPlacement, allowlisted_placements)
  mapper(settings.MccIds, mcc_ids)
  mapper(settings.CustomerIds, customer_ids)
  if engine:
    metadata.create_all(engine)
